/*
 * kissdx - KiSS PC-Link Daemon eXtended (based on kissd)
 *
 * Copyright (C) 2005 Stelian Pop <stelian@popies.net>
 * Portions Copyright (C) 2006 Vidar Tysse <vidar@vidartysse.net>
 *
 * Heavily based on kiss4lin,
 * Copyright (C) 2004 Jacob Kolding <dacobi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * ************************************************************************
 *
 * utils.c
 * 
 * Purpose:		A collection of general-purpose utilities in the following
 * 				categories:
 * 				-	String handling, including list handling and substitution
 * 				-	Character set conversion
 * 				-	Case-insensitive filename operations
 * 
 * ************************************************************************
 */

// We use iconv by default.
// On platforms that don't have it we must patch this file to remove the #define below.
#define USE_ICONV

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <dirent.h>
#include <libgen.h>

#ifdef USE_ICONV
#include <iconv.h>
#endif

#include "utils.h"
#include "kissdx.h"


/* ************************
 * Character set conversion
 * ************************
 */

const char* convert_charset(const char* s, enum convertdirection_t direction)
{
#ifdef USE_ICONV
	static char outbuf[2 * PATH_MAX];
	char *wrptr = (char*)outbuf;
	size_t inbytesleft = strlen(s), outbytesleft = 2 * PATH_MAX - 1;
	static int isOpenErrReported = 0, isConvErrReported = 0;

	if (!*client_character_set || !*server_character_set)
		return s;			/* Conversion not configured: Return original */

	memset(outbuf, 0, sizeof outbuf);

	char *to_character_set   = direction == cdt_to_client? client_character_set : server_character_set;
	char *from_character_set = direction == cdt_to_client? server_character_set : client_character_set;

    iconv_t cd = iconv_open(to_character_set, from_character_set);
    if (cd == (iconv_t) -1) {
    	if (!isOpenErrReported) {
    		log("iconv_open fails for '%s' --> '%s' conversion: %s", from_character_set, to_character_set, strerror(errno));
    		isOpenErrReported = -1;
    	}
    	return s;			/* Cannot convert: Return original */
    }

	const char *ps = s;
	size_t nconv = iconv(cd, &ps, &inbytesleft, &wrptr, &outbytesleft);
	if (nconv == (size_t) -1) {
    	if (!isConvErrReported) {
    		log("iconv fails for '%s' --> '%s' conversion at '%s': %s", from_character_set, to_character_set, ps, strerror(errno));
    		isConvErrReported = -1;
    	}
    	iconv_close(cd);
    	return s;			/* Cannot convert: Return original */
    }

	if (iconv_close(cd) != 0) {
		log("iconv_close failed for '%s' --> '%s' conversion: %s", from_character_set, to_character_set, strerror(errno));
    	return s;			/* Cannot convert: Return original */
    }

	return (char*)outbuf;
#else
	static int isConvErrReported = 0;
	if (!isConvErrReported) {
		log0("iconv support has not been compiled in, probably because it is not supported on this platform!");
		isConvErrReported = -1;
	}
	return s;			/* Conversion not possible: Return original */
#endif
}


/* ****************************************
 * String handling, including list handling
 * ****************************************
 */


// Make contents of buf lower-case
void make_lowercase(char *buf)
{
	static const char *upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	static const char *lower = "abcdefghijklmnopqrstuvwxyz";
	char *p, *q;
	for (p = buf; *p != '\0'; p++)
		if ((q = strchr(upper, *p)))
			*p = lower[q - upper];
}


// Return a pointer to the first occurrence of needle in haystack
const char* list_pointer(const char *haystack, const char *needle, const char *delimiters)
{
	const char *p = haystack, *q;
	if (haystack && needle && delimiters && *haystack && *needle && *delimiters) {
		int needle_len = strlen(needle);
		while ((p = strstr(p, needle))) {
			if (p == haystack || strchr(delimiters, *(p-1)))
				if (*(q = p + needle_len) == '\0' || strchr(delimiters, *q))
					return p;
			p += needle_len;
		}
	}
	return 0;
}


// Return true if haystack contains needle, false otherwise
int list_contains(const char *haystack, const char *needle, const char *delimiters)
{
	return list_pointer(haystack, needle, delimiters) != NULL;
}


// Return true if comma-delimited list contains filename's extension, false otherwise
int list_contains_extension(const char *list, const char *filename)
{
	const char *p = strrchr(filename, '.');
	if (p) {
		char extLower[32];
		memset(extLower, 0, sizeof(extLower));
		strncpy(extLower, p + 1, sizeof(extLower));
		*(extLower + sizeof(extLower) - 1) = '\0';
		make_lowercase(extLower);
		return list_contains(list, extLower, ",");
	}
	return 0;
}


// Substitute every occurrence of needle in haystack with substitute
int string_substitute(char *haystack, const char *needle, const char *substitute, int haystack_size)
{
	int num_substitutions = 0;
	if (haystack && needle && substitute && *haystack && *needle && *substitute && haystack_size > strlen(haystack)) {
		char *p = haystack;
		int needle_len = strlen(needle);
		int substitute_len = strlen(substitute);
		int shift_amount = substitute_len - needle_len;
		haystack_size--;					// Ensure room for trailing null char
		while ((p = strstr(p, needle))) {
			int chars_free = haystack_size - strlen(haystack);
			if (chars_free < shift_amount)
				return num_substitutions;	// Stop if overflow
			memmove(p + substitute_len, p + needle_len, strlen(p + needle_len) + 1);	// Make exact room for substitute
			memcpy(p, substitute, substitute_len);										// Overlay substitute
			num_substitutions++;
			p += substitute_len;
		}
	}
	return num_substitutions;
}


/* ************************************
 * Case-insensitive filename operations
 * ************************************
 */


/// A case-insensitive stat()
/// NOTE 1:	Only the base name of filename is compared case-insensitively, the directory
///			part of the supplied filename must be in correct case.
/// NOTE 2:	In addition to normal stat() behaviour (though case-insensitive), this function
///			places the actual filename of the file found into the supplied filename buffer,
///			which is thus not const char * but char * for this implementation.
int casestat(char *filename, struct stat *statbuf)
{
	char directory[PATH_MAX], name[PATH_MAX];

	int scandir_selector(const struct dirent *entry)
	{
		return !strcasecmp(name, entry->d_name);
	}

	char fullname[PATH_MAX], foundname[PATH_MAX];
	struct dirent **namelist;
	int num_files, n;

	if (!filename)
		return -1;

	strncpy(fullname, filename, sizeof fullname);
	if (*fullname != '/') {
		if (!getcwd(fullname, sizeof fullname)) {
			log("casestat: getcwd: %s", strerror(errno));
			return -1;
		}
		if (*(fullname + strlen(fullname) - 1) != '/')
			strcat(fullname, "/");
		strcat(fullname, filename);
	}
	strcpy(name, basename(fullname));
	strcpy(directory, dirname(fullname));

	if ((num_files = scandir(directory, &namelist, scandir_selector, alphasort)) < 0) {
		log("casestat: scandir '%s': %s", directory, strerror(errno));
		return -1;
	}
	for (*foundname = '\0', n = 0; n < num_files; n++) {
		if (!*foundname)
			strcpy(foundname, namelist[n]->d_name);
		free(namelist[n]);
	}
	free(namelist);

	if (!*foundname)
		return -1;

	strcpy(fullname, directory);
	strcat(fullname, "/");
	strcat(fullname, foundname);
	if (stat(fullname, statbuf) < 0) {
		log("casestat: stat '%s': %s", fullname, strerror(errno));
		return -1;
	}
	
	strcpy(filename, fullname);		// We send back the actual (correct-case) filename to the caller
	return 0;
}
